"""
The parent module of spyral-utils

Modules
-------
constants
    Useful constants for AT-TPC
plot
    Plotting utilities such as Cut2D and Histogrammer
nuclear
    Nuclear physics tools like nuclear masses, momentum 4-vectors, energy loss, etc.
"""
